##- Nanar <nanardon@mandrake.org>
##-
##- This program is free software; you can redistribute it and/or modify
##- it under the terms of the GNU General Public License as published by
##- the Free Software Foundation; either version 2, or (at your option)
##- any later version.
##-
##- This program is distributed in the hope that it will be useful,
##- but WITHOUT ANY WARRANTY; without even the implied warranty of
##- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##- GNU General Public License for more details.
##-
##- You should have received a copy of the GNU General Public License
##- along with this program; if not, write to the Free Software
##- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# $Id: Solve.pm,v 1.1 2005/04/19 23:08:32 othauvin Exp $

package Hdlist::Media::Solve;

use strict;
use warnings;
use Hdlist::Media;
use Hdlist::Header;

sub new {
    my ($class, %options) = @_;

    my $solve = {
        db => Hdlist::newdb($options{root}),
        medium => [],
        sysdep => {
            'REQUIRE' => undef,
            'CONFLICT' => undef,
        },
        request => {},
        trans => {},
        
        long_fmt => '%{NAME}-%{VERSION}-%{RELEASE}.%{ARCH} (%{SUMMARY})',
        short_fmt => '%{NAME}-%{VERSION}-%{RELEASE}',

        update_all_dep => 0, # if set, try to update every need packages

        choice_install => sub {
            my ($pkgstate, $pkg, $dep, @list) = @_;
            printf("To %s %s%s, I need one of this:\n%s\nYour choice ? ",
                $pkgstate,
                $pkg || "",
                $dep ? " [ $dep ]" : "",
                join("\n", @list),
            );
            print "1 (auto-selecting)\n";
            1;
        },        
                
        confirm_remove => sub {
            my ($pkgstate, $pkg, @list) = @_;
            printf("To %s %s, I have to remove:\n%s\nDo you agree ? ",
                $pkgstate,
                $pkg || "",
                join("\n", @list),
            );
            print "Y (auto-reply)\n";
            1;
        },

        info_notremove => sub {
            my (@list) = @_;
            printf("Can't remove those pkg:\n%s\n",
                join("\n", @list),
            );
            1;
        },

        info_notinstall => sub {
            my ($pkg, $dep) = @_;
            printf("Can't install %s (missing %s)\n",
                $pkg,
                $dep,
            );
            1;
        },
    };

    foreach (keys %{$solve}) {
        $solve->{$_} = $options{$_} if $options{$_};
    }
    
    bless($solve, $class);
}

sub addmedia {
    my ($self, $media) = @_;
    push(@{$self->{medium}}, $media);
    $#{$self->{medium}};
}

sub load {
    my ($self, $callback) = @_;

    if ($self->{sysdep}{REQUIRE}) {
        $self->{sysdep}{REQUIRE}->init();
        while ($self->{sysdep}{REQUIRE}->next >= 0) {
            substr($self->{sysdep}{REQUIRE}->name, 0, 1) eq '/' and $self->{reqfiles}{$self->{sysdep}{REQUIRE}->name} = 1;
        }
    }
    
    $self->{db}->traverse(sub {
            my ($h) = @_;
            @{$self->{reqfiles}}{$h->requiredfiles()} = (); 
            1; }
    ) if (@{$self->{medium}});
    foreach my $source (@{$self->{medium}}) {
        $source->init(sub {
                my (%arg) = @_;
                $arg{header}->archscore or return 0;
                $arg{header}->osscore or return 0;
                @{$self->{reqfiles}}{$arg{header}->requiredfiles()} = ();
                1; }
        );
    }
    foreach my $source (@{$self->{medium}}) {
        $source->load([ keys %{$self->{reqfiles}} ]);
    }

    my $pname = Hdlist::flagvalue("dbquery", "PROVIDENAME");
    my $rname = Hdlist::flagvalue("dbquery", "REQUIRENAME");
    $self->traverse(
        db => 0,
        medium => undef,
        callback => sub {
            my ($h, $i, $m) = @_;

            $self->add_in_cache('NAME', $h->tag('NAME'), $m , $i);

            if (my $f = $h->files) {
                $f->init;
                while ($f->next >= 0) {
                    $self->add_in_cache('BASENAMES', $f->filename, $m , $i);
                }
            }
            
            foreach (qw(PROVIDENAME REQUIRENAME OBSOLETENAME CONFLICTNAME)) {
                my $tag = Hdlist::flagvalue("dbquery", $_);
                if (my $d = $h->dep($tag)) {
                    $d->init;
                    while ($d->next >= 0) {
                        $self->add_in_cache($tag, $d->name, $m , $i);
                    }
                }
            }
            1;
        },
    );
}

sub add_system_dep {
    my ($self, $tag, $name, $sense, $evr) = @_;
    if (defined($self->{sysdep}{$tag})) {
        $self->{sysdep}{$tag}->add($name, $sense, $evr);
    } else {
        $self->{sysdep}{$tag} = Hdlist::newdep($tag, $name, $sense || 0, $evr || '');
    }
}

sub add_request {
    my ($self, $media, $id) = @_;
    $self->{request}{$media}{$id} ++;
}

sub clean_request {
    my ($self) = @_;
    $self->{request} = undef;
}

sub newtransid {
    my ($self) = @_;
    ++$self->{transid}
}

sub add_trans_set {
    my ($self, $trans, $media, $id) = @_;
    my $newtrans = undef;
    my $needby = defined($trans) ? "$trans->{media}:$trans->{id}" : 'Request';
    my $transid = defined($trans) ? $trans->{transid} : $self->newtransid();
    if (!exists($self->{trans}{$media}{$id})) {
        $newtrans = {
            transid => $transid,
            id => $id,
            media => $media,
            header => $self->get_header($media, $id),
            goal => $trans->{goal} || ($media eq 'DB' ? '-' : '+'),
        };
    }
    print STDERR "add_trans_set(): Adding new trans $media:$id ($transid $needby)\n";
    if (!defined($newtrans->{header})) { return $newtrans };
    $self->{trans}{$media}{$id}{$transid} = 1;
    $self->{trans_tree}{$transid}{$needby}{"$media:$id"} = 1;
    $newtrans
}

sub rollback_trans {
    my ($self, $trans) = @_;
    print STDERR "rollback_trans(): $trans->{transid}\n";
    foreach my $media (keys %{$self->{trans} || {}}) {
        foreach my $id (keys %{$self->{trans}{$media} || {}}) {
            delete $self->{trans}{$media}{$id}{$trans->{transid}};
            delete $self->{trans}{$media}{$id} if (! keys %{$self->{trans}{$media}{$id}});
        }
    }
    delete $self->{trans_tree}{$trans->{transid}};
}

sub clean_trans {
    my ($self) = @_;
    $self->{trans_tree} = undef;
    $self->{trans} = undef;
}

sub add_in_cache {
    my ($self, $tag, $name, $m, $i) = @_;
    my $tagv = Hdlist::flagvalue("dbquery", $tag);
    $self->{incache}{$tagv} = 1;
    grep($i, @{$self->{cache}{$name}{$tagv}{$m} || []}) and return;
    push (@{$self->{cache}{$name}{$tagv}{$m}}, $i);
}

sub traverse_medium {
    my ($self, %options) = @_;
    defined($options{tag}) and $options{tag} = Hdlist::flagvalue("dbquery", $options{tag});

    my $continue = 1;
    if (defined($options{tag})) {
        # using cache
        exists $self->{cache}{$options{tagvalue}} or return;
        foreach my $m (keys %{$self->{cache}{$options{tagvalue}}{$options{tag}} || {}}) {
            foreach my $i (@{$self->{cache}{$options{tagvalue}}{$options{tag}}{$m}}) {
                next if ($options{virtual} && !exists($self->{trans}{$m}{$i}));
                $continue = $options{callback}->(
                    $self->get_header($m, $i),
                    $i,
                    $m,
                ) or return 0;
            }
        }
    } else {
        foreach my $m (0 .. $#{$self->{medium}}) {
            $self->{medium}[$m]->traverse(
                sub {
                    my ($h, $i) = @_;
                    return $continue = $options{callback}->($h, $i, $m);
                },
            );
        }
    }
    return $continue;
}

sub traverse {
    my ($self, %options) = @_;
    my $continue = 1;

if (1) {
    if ($options{db}) {
        $self->{db}->traverse(
            sub {
                $continue = $options{callback}->(
                    $_[0],
                    $_[1],
                    'DB',
                );
                $continue;
            },
            $options{tag}, 
            $options{tagvalue},
        );
    }
    $continue or return;

}
    $self->traverse_medium(%options);
}

sub traverse_virtual {
    my ($self, %options) = @_;
    #printf(STDERR "traverse_virtual start %s=%s\n", $options{tag} || '', $options{tagvalue} || '');

    $self->traverse_medium(%options, virtual => 1) or return;

    $self->{db}->traverse(
        sub {
            exists($self->{trans}{'DB'}{$_[1]}) and return 1;
            $options{callback}->(
                $_[0],
                $_[1],
                'DB',
            );
        },
        $options{tag}, 
        $options{tagvalue},
        0,
        [ keys %{$self->{trans}{'DB'} || {}} ],
    );
}

sub get_header {
    my ($self, $media, $id) = @_;
    if ($media eq 'DB') {
        return $self->{db}->get_header($id);
    } elsif (defined($self->{medium}[$media])) {
        return $self->{medium}[$media]->get_header($id);
    }
}

sub list_trans {
    my ($self) = @_;
    my $tree;
    foreach my $t (keys %{$self->{trans_tree} || {}}) {
        foreach my $n (keys %{$self->{trans_tree}{$t} || {}}) {
            foreach my $m (keys %{$self->{trans_tree}{$t}{$n} || {}}) {
                $tree->{$m}{$n} = 1;
            }
        }
    }

    foreach my $media (keys %{$self->{trans}}) {
        foreach my $id (keys %{$self->{trans}{$media}}) {
            printf "[%s] %s (%s)\n", $media eq 'DB' ? '-' : '+', scalar($self->get_header($media, $id)->fullname()), join(", ", map { $_ ne 'Request' and $_ = scalar($self->get_header(split(':', $_))->fullname); $_ } keys %{$tree->{"$media:$id"}});
        }
    }
}

# Check this require is in trans else callback
sub check_require {
    my ($self, $require, $callback) = @_;
    my $foundheader = undef;
    if (substr($require->name(), 0, 1) eq '/') {
        # is a file
        $self->traverse_virtual(
            tag => 'BASENAMES',
            tagvalue => $require->name(),
            callback => sub {
                my ($i, $m);
                ($foundheader, $i, $m) = @_;
                exists($self->{request}{$m}{$i}) ? 1 : 0; # stop search if found
            },
        );
    } else {
        # is a std requirement
        $self->traverse_virtual(
            tag => 'PROVIDENAME',
            tagvalue => $require->name(),
            callback => sub {
                my ($h, $i, $m) = @_;
                if ($h->matchdep($require)) {
                    $foundheader = $h;
                    return exists($self->{request}{$m}{$i}) ? 1 : 0; # Stop traverse
                } else {
                    return 1;
                }
            }
        );
    }
    if (defined($foundheader)) {
        if (!$self->{update_all_dep} || !defined($callback)) {
            return 1;
        }
    } else {
        printf(STDERR "%s not found in transaction\n", scalar($require->info()));
        defined($callback) or return 0;
    }
    my @solve_solutions;
    if (substr($require->name(), 0, 1) eq '/') {
        $self->traverse(
            tag => 'BASENAMES',
            tagvalue => $require->name(),
            callback => sub {
                my ($h, $id, $media) = @_;
                push @solve_solutions, [ $h, $media, $id ];
                1;
            },
        );
    } else {
        $self->traverse(
            db => 0,
            medium => undef,
            tag => 'PROVIDENAME',
            tagvalue => $require->name(),
            callback => sub {
                my ($h, $id, $media) = @_;
                $h->matchdep($require) and do {
                    push @solve_solutions, [ $h, $media, $id ];
                    1;
                };
            },
        );
    }
    if (defined($foundheader)) {
        @solve_solutions = grep { !exists($self->{trans}{$_->[1]}{$_->[2]}) && $_->[0]->is_better_than($foundheader) } @solve_solutions;
        scalar(@solve_solutions) or do { return 1; };
        $callback->($require, @solve_solutions);
        return 1;
    } else {
        @solve_solutions or return 0;
        return $callback->($require, @solve_solutions);
    }
}

sub find_update {
    my ($self, $header, $callback) = @_;
    my @find;
    $self->traverse(
        db => 0,
        medium => undef,
        tag => 'NAME',
        tagvalue => $header->tag('NAME'),
        callback => sub {
            my ($h, $i, $m) = @_;
            $h->compare($header) and push @find, [$h, $m, $i];
            1;
        },
    );
    $self->traverse(
        db => 0,
        medium => undef,
        tag => 'OBSOLETENAME',
        tagvalue => $header->tag('NAME'),
        callback => sub {
            my ($h, $i, $m) = @_;
            if (my $obs = $h->dep('O')) {
                $obs->init;
                while ($obs->next >= 0) {
                    $obs->matchheadername($header) and do {
                        push @find, [$h, $m, $i];
                        last;
                    };
                }
            }
            1;
        },
    );
    @find or return 0;
    return $callback->(undef, @find);
}

sub check_requires {
    my ($self, $header, $callback) = @_;
    printf STDERR "Check requires %s\n", scalar($header->fullname());
    my $requires = $header->dep('REQUIRE') or return 1;
    $requires->init();
    while ($requires->next() != -1) {
        my $res = 0;
        $requires->flags & 1 << 24 and next;
        $self->check_require($requires, $callback) or do {
            $self->{info_notinstall}->(
                $header->queryformat($self->{short_fmt}),
                scalar($requires->info),
            );
            return 0;
        };
    }
    return 1;
}

sub check_provides {
    my ($self, $header, $callback, $callbackrequires) = @_;
    $callback ||= sub { 0 };
    printf STDERR "Check provides %s\n", scalar($header->fullname());
    if (my $provides = $header->dep('PROVIDE'))  {
        $provides->init();
        while ($provides->next != -1) {

            if ($self->{sysdep}{REQUIRE}) {
                $self->{sysdep}{REQUIRE}->init();
                while ($self->{sysdep}{REQUIRE}->next >= 0) {
                    if ($self->{sysdep}{REQUIRE}->overlap($provides)) {
                        $self->check_require($self->{sysdep}{REQUIRE}, $callbackrequires) or do {
                            printf("Can't remove %s due to solve dep %s\n",
                                scalar($header->fullname()),
                                scalar($self->{sysdep}{REQUIRE}->info()));
                            return 0;
                        };
                    }
                }
            }
            
            my $res = 0;
            $self->traverse_virtual(
                tag => 'REQUIRENAME',
                tagvalue => $provides->name(),
                callback => sub {
                    my ($h, $id, $media) = @_;
                    my $requires = $h->dep('REQUIRE') or return 1;
                    $requires->init();
                    while ($requires->next() != -1) {
                        $requires->name() eq $provides->name() or next;
                        $self->check_require($requires, $callbackrequires) or do {
                            $callback->($requires, $h, $media, $id) or do {
                                $res ++;
                                return 0;
                            };
                            last;
                        };
                    }
                    1;
                },
            );
            $res and return 0;
        }
    }
    if (my $files = $header->files()) {
        $files->init();
        while ($files->next() >= 0) {
            defined($self->{reqfiles}{$files->filename}) or next;
            my $res = 0;
            $self->traverse_virtual(
                tag => 'REQUIRENAME',
                tagvalue => $files->filename(),
                callback => sub {
                    my ($h, $id, $media) = @_;
                    my $requires = $h->dep('REQUIRE') or return 1;
                    $requires->init();
                    while ($requires->next() >= 0) {
                        $requires->name() eq $files->filename() or next;
                        $self->check_require($requires, $callbackrequires) or do {
                            $callback->($requires, $h, $media, $id) or do {
                                $res ++;
                                return 0;
                            };
                        };
                    }
                    1;
                },
            );
            $res and return 0;
        }
    }
    return 1;
}

sub check_conflicts {
    my ($self, $header, $callback) = @_;
    $callback ||= sub { 0 };
    printf STDERR "Check conflicts %s\n", scalar($header->fullname());
    my $res = 0;
    if ($self->{sysdep}{CONFLICT}) {
        $self->{sysdep}{CONFLICT}->init();
        while ($self->{sysdep}{CONFLICT}->next >= 0) {
            $self->{sysdep}{CONFLICT}->matchheader($header) and do {
                return 0;
            };
        }
    }
    
    if (my $conflicts = $header->dep('CONFLICT')) {
        $conflicts->init();
        while ($conflicts->next >= 0) {
            $self->traverse_virtual(
                tag => 'PROVIDENAME',
                tagvalue => $conflicts->name,
                callback => sub {
                    my ($h, $i, $m) = @_;
                    if ($h->matchdep($conflicts)) {
                        $callback->($conflicts, $h, $m, $i) or do {
                            $res = 1;
                            return 0;
                        };
                    }
                    1;
                },
            );
        }
    }
    $res and return 0;
    if (my $provides = $header->dep('PROVIDE')) {
        $provides->init();
        while ($provides->next >= 0) {
            $self->traverse_virtual(
                tag => 'CONFLICTNAME',
                tagvalue => $provides->name,
                callback => sub {
                    my ($h, $i, $m) = @_;
                    if (my $c = $h->dep('C')) {
                        $c->init();
                        while ($c->next >= 0) {
                            $c->name eq $provides->name or next; # Don't bother
                            if ($header->matchdep($c)) {
                                $callback->($c, $h, $m, $i) or do {
                                    $res = 1;
                                    return 0;
                                };
                            }
                        }
                    }
                    1;
                },
            );
        }
    }
    $res and return 0;
    1;
}

sub find_old {
    my ($self, $header, $callback) = @_;
    $callback ||= sub { 0 };
    printf STDERR "Check old %s\n", scalar($header->fullname());
    my $res = 0;
    $self->traverse_virtual(
        tag => 'NAME',
        tagvalue => $header->tag('NAME'),
        callback => sub {
            my ($h, $id, $media) = @_;
            $header eq $h and return 1;
            #if ($header->compare($h) == 0) {
                #printf ("%s is allready installed\n", scalar($header->fullname));
                #$res++;
                #    return 1;
            #}
            $callback->(undef, $h, $media, $id) or do {
                $res++;
                return 0;
            };
            1;
        },
    );
    return $res ? 0 : 1;
}

sub find_obsoleted {
    my ($self, $header, $callback) = @_;
    $callback ||= sub { 0 };
    printf STDERR "Check obsoletes %s\n", scalar($header->fullname());
    my $obsoletes = $header->dep('OBSOLETE') or return 1;
    $obsoletes->init();
    my $res = 0;
    while ($obsoletes->next != -1) {
        $self->traverse_virtual(
            tag => 'NAME',
            tagvalue => $obsoletes->name(),
            callback => sub {
                my ($h, $id, $media) = @_;
                $header eq $h and return 1;
                if ($h->namematchdep($obsoletes)) {
                    $callback->("", $h, $media, $id) or do {
                        $res++;
                        return 0;
                    };
                }
                1;
            },
        );
    }
    $res ? 0 : 1;
}

sub autoselect {
    my ($self) = @_;

    $self->traverse_virtual(
        callback => sub {
            my ($header, $id, $media) = @_;
            $self->find_update($header,
                sub {
                    my (undef, @list) = @_;
                    my ($s) = $self->sort_proposal(@list);
                    print $s->[0] ."\n";
                    defined($s->[0]) or return;
                    $self->{trans}{$s->[1]}{$s->[2]}{'AUTO'} = 1;
                    printf "auto: %s => %s\n", scalar($header->fullname), scalar($s->[0]->fullname);
                }
            );
            1;
        }
    );

    foreach my $m (keys %{$self->{trans} || {}}) {
        foreach my $i (keys %{$self->{trans}{$m} || {}}) {
            exists($self->{trans}{$m}{$i}{'AUTO'}) or next;
            $self->add_request($m, $i);
        }
    }
    $self->rollback_trans({ transid => 'AUTO' });
}

sub sort_proposal {
    my ($self, @list) = @_;
    foreach my $item (@list) {
        exists($self->{trans}{$item->[1]}{$item->[2]}) and return($item);
    }
    @list = sort { $b->[0]->is_better_than($a->[0]) } @list;
}


sub check_transaction {
    my ($self, $trans, $type) = @_;

    my $possible = 1;

    my $sub_auto_remove = sub {
        my ($dep, $h, $m, $i) = @_;
        printf(STDERR "Removing %s [%s]\n",
            $h->queryformat($self->{short_fmt}),
            $dep ? scalar($dep->info) : "",
        );
        return $self->add_transaction($m, $i, '-', $trans)
    };
    
    my @toremoved = ();
    my $sub_confirme_remove = sub {
        my ($dep, $h, $m, $i) = @_;
        push(@toremoved, [
            sprintf("%s%s",
                $h->queryformat($self->{short_fmt}),
                $dep ? " [" . $dep->info . "]" : "",
            ),
            $m,
            $i
            ]
        );
        1;
    };
    
    my %addchoice;
    my $subadd = sub {
        my ($dep, @list) = @_;
        @list = grep { $_->[1] ne 'DB' } @list;
        scalar(@list) or return 0;
        if (@list > 1) {
            @{$addchoice{$dep ? scalar($dep->info) : "upd"}} = @list;
            return 1;
        } else {
            printf STDERR ($dep ? $dep->info : ""). " + $list[0][1]:$list[0][2] for ". $trans->{header}->fullname."\n";
            return $self->add_transaction($list[0][1], $list[0][2], '+', $trans);
        }
    };

    my $subaddsel = sub {
        foreach my $d (keys %addchoice) {
            @{$addchoice{$d}} = $self->sort_proposal(@{$addchoice{$d}});
            my $res = @{$addchoice{$d}} > 1 ? $self->{choice_install}->(
                "install",
                $trans->{header}->queryformat($self->{short_fmt}),
                $d,
                map { $_->[0]->queryformat($self->{long_fmt}) } @{$addchoice{$d}},
            ) : 1;
            return 0 if ($res < 1 || $res > $#{$addchoice{$d}});
            $res--;
            $self->add_transaction($addchoice{$d}[$res][1], $addchoice{$d}[$res][2], '+', $trans) or return 0;
        }
        1;
    };

    if ($type eq '+') {
        if ($possible) {
            $self->find_old($trans->{header}, $sub_auto_remove) or $possible = 0;
        }
        if ($possible) {
            $self->find_obsoleted($trans->{header}, $sub_auto_remove) or $possible = 0;
        }
        if ($possible) {
            $self->check_requires($trans->{header}, $subadd) or $possible = 0;
            $subaddsel->() or $possible = 0;
        }
        if ($possible) {
            $self->check_conflicts($trans->{header}, $sub_confirme_remove) or $possible = 0;
            if (@toremoved) {
                if ($possible = $self->{confirm_remove}->(
                    "install",
                    $trans->{header}->queryformat($self->{short_fmt}),
                    map { $_->[0] } @toremoved,
                )) {
                    foreach (@toremoved) {
                        $self->add_transaction($_->[1], $_->[2], '-', $trans) or do {
                            $possible = 0;
                            last;
                        };
                    }
                }
            }
        }
    } else {
        if ($possible) {
            $self->check_provides(
                $trans->{header}, 
                $trans->{goal} eq '+' ? sub { $self->find_update($_[1], $subadd) || $sub_auto_remove->(@_) } : $sub_auto_remove,
                $trans->{goal} eq '+' ? $subadd : undef,
            )  or $possible = 0;
        }
    }

    print STDERR $type . " " . scalar($trans->{header}->fullname) . " Is possible $possible\n";
    return $possible;
}

sub add_transaction {
    my ($self, $media, $id, $type, $parenttrans) = @_;
    printf STDERR "Add $type $media:$id %s\n", join(" ", caller);
    ($type eq '+' && $media eq 'DB') || ($type eq '-' && $media ne 'DB') and do {
        print STDERR "I can't undo this $type $media $id\n";
        return 0;
    };
    my $trans;
    my $res;
    $trans = $self->add_trans_set($parenttrans, $media, $id);
    defined($trans->{header}) or return 1;
    
    printf STDERR "Check trans %s %s:%s %s\n", $type, $media, $id, scalar($trans->{header}->fullname);

    $res = $self->check_transaction($trans, $type);

    if ($res) {
        return 1;
    } else {
        delete $self->{trans}{$media}{$id}{$trans->{transid}};
        $self->rollback_trans($trans) if(!defined($parenttrans));
        return 0;
    }
}

sub check {
    my ($self) = @_;

    foreach my $media (keys %{$self->{request}}) {
        foreach my $id (keys %{$self->{request}{$media}}) {
            $self->{request}{$media}{$id} = 0;
        }
    }

    my $run = 1;
    while ($run) {
        $run = 0;
        foreach my $media (keys %{$self->{request}}) {
            $media eq 'DB' and next;
            foreach my $id (keys %{$self->{request}{$media}}) {
                $self->{request}{$media}{$id} and next;
                $self->{request}{$media}{$id} = 1;
                $run = 1;
                $self->add_transaction($media, $id, '+');
            }
        }

        foreach my $id (keys %{$self->{request}{'DB'} || {}}) {
            $self->{request}{'DB'}{$id} and next;
            $self->{request}{'DB'}{$id} = 1;
            $run = 1;
            $self->add_transaction('DB', $id, '-');
        }
    }

    foreach my $m (keys %{$self->{trans}}) {
        foreach my $i (keys %{$self->{trans}{$m} || {}}) {
            if ($m eq 'DB') {
                $self->{db}->transremove($i);
            } else {
                $self->{db}->transadd(
                    $self->get_header($m, $i),
                    "$m:$i"
                );
            }
        }
    }

    
    $self->{db}->transcheck();
    my @r = Hdlist::format_rpmpb($self->{db}->transpb());
    foreach (@r) {
        print "$_\n";
    }
    $self->{db}->transclean();
    $self->{db}->transreset();
}

sub run {
    my ($self) = @_;

    my $trans;

    foreach my $m (keys %{$self->{trans}}) {
        foreach my $i (keys %{$self->{trans}{$m} || {}}) {
            my ($transid) = sort { $a <=> $b } keys %{$self->{trans}{$m}{$i}};
            push @{$trans->{$transid}}, [ $m, $i ];
        }
    }

    foreach my $transid (keys %{$trans}) {
        foreach my $item (@{$trans->{$transid}}) {
            if ($item->[0] eq 'DB') {
                $self->{db}->transremove($item->[1]);
                print STDERR "Removing $item->[1]\n";
            } else {
                print $self->{medium}[$item->[0]]->get_rpm($self->{medium}[$item->[0]]->id2rpm($item->[1])) ."\n";
                my $h = Hdlist::rpm2header($self->{medium}[$item->[0]]->get_rpm($self->{medium}[$item->[0]]->id2rpm($item->[1])));
                $self->{db}->transadd(
                    $h,
                    $self->{medium}[$item->[0]]->get_rpm($self->{medium}[$item->[0]]->id2rpm($item->[1])),
                );
            }
        }
        
        $self->{db}->transcheck();
        my @r = Hdlist::format_rpmpb($self->{db}->transpb());
        foreach (@r) {
            print "$_\n";
        }
        if (!@r) { 

            $self->{db}->transflag([ ]);
            #$self->{db}->transclean();
            $self->{db}->transorder();
            $self->{db}->transrun(undef);
        }
        $self->{db}->transclean();
        $self->{db}->transreset();
    }
    $self->clean_trans(); 
    $self->clean_request();
}


#### USERLAND

sub add_from_require {
    my ($self, $name, $flags, $evr) = @_;

    my $subadd = sub {
        my ($dep, @list) = @_;
        my $res = 0;
        if (@list > 1) {
            $self->{choice_install}->(
               "Request", "",
               scalar($dep->info),
               map { $_->[0]->queryformat($self->{long_fmt}); } @list
            );
        } else {
            $res = 1;
            printf("Find for %s package %s\n",
               scalar($dep->info),
               $list[0][0]->queryformat($self->{short_fmt}),
            );
        }
        $res < 1 || $res > scalar(@list) and return 0;
        $res--;
        return $self->add_transaction($list[$res][1], $list[$res][2], '+', undef);
    };
    
    my $r = Hdlist::newdep('R', $name, $flags, $evr);
    $r->init; $r->next;

    $self->check_require($r, $subadd) or print "Can't find this dep\n";
    
}

sub find_and_add {
    my ($self, $name, %options) = @_;

    $self->traverse(
        callback => sub {
            my ($header, $id, $media) = @_;
            $header->tag(1000) eq $name and do {
                $self->add_request($media, $id);
                print $header->fullname ."\n";
                return 0;
            };
            1;
        },
    );
}

sub find_and_remove {
    my ($self, $name, %options) = @_;
    $self->traverse(
        db => 1,
        medium => [],
        tag => 'NAME',
        tagvalue => $name,
        callback => sub {
            my ($header, $off) = @_;
            $header->tag(1000) eq $name and do {
                print $header->fullname . " " . $off ."\n";
                $self->add_request('DB', $off);
                return 0;
            };
            1;
        },
    );
}

1;
